"""
Pure-Python core library for FluidSim framework. This package provides
generic base classes and utilities to build new solvers.

.. autosummary::
   :toctree:

   params
   info
   solver
   output
   loader
   magic
   extend_simul
   hexa_files
   scripts
   paths

"""

from ._version import __version__
